@@ -1,7 +1,7 @@
 /*
  * RS-PostgreSQL.c
  *
- * $Id: RS-PostgreSQL.c 237 2012-10-04 14:54:29Z tomoakin@kenroku.kanazawa-u.ac.jp $
+ * $Id$
  *
  * This package was developed as a part of Summer of Code program organized by Google.
  * Thanks to David A. James & Saikat DebRoy, the authors of RMySQL package.
@@ -71,6 +71,89 @@ struct data_types RS_PostgreSQL_dataTypes[] = {
  */
 
 
+/* RS_PostgreSQL_closeManager */
+/* s_object *
+ * RS_PostgreSQL_closeManager(Mgr_Handle * mgrHandle)
+ * {
+ *     S_EVALUATOR RS_DBI_manager * mgr;
+ *     s_object *status;
+ */
+
+s_object * RS_PostgreSQL_closeManager(Mgr_Handle * mgrHandle);
+
+/* 
+ * RS_DBI_SclassNames
+ * RS_DBI_validHandle
+ * RS_PostgreSQL_cloneConnection
+ * RS_PostgreSQL_closeConnection
+ * RS_PostgreSQL_closeManager
+ * RS_PostgreSQL_closeResultSet
+ * RS_PostgreSQL_connectionInfo
+ * RS_PostgreSQL_CopyIn
+ * RS_PostgreSQL_CopyInDataframe
+ * RS_PostgreSQL_dbApply
+ * RS_PostgreSQL_escape
+ * RS_PostgreSQL_escape_bytea
+ * RS_PostgreSQL_exec
+ * RS_PostgreSQL_fetch
+ * RS_PostgreSQL_getException
+ * RS_PostgreSQL_getResult
+ * RS_PostgreSQL_init
+ * RS_PostgreSQL_managerInfo
+ * RS_PostgreSQL_newConnection
+ * RS_PostgreSQL_pqexec
+ * RS_PostgreSQL_resultSetInfo
+ * RS_PostgreSQL_typeNames
+ * RS_PostgreSQL_unescape_bytea
+ */
+
+static const R_CMethodDef cMethods[] = {
+   {NULL, NULL, 0, NULL}
+};
+
+static const R_CallMethodDef callMethods[]  = {
+  {"RS_DBI_SclassNames", (DL_FUNC) &RS_DBI_SclassNames, 1},
+  {"RS_DBI_validHandle", (DL_FUNC) &RS_DBI_validHandle, 1},
+  {"RS_PostgreSQL_cloneConnection", (DL_FUNC) &RS_PostgreSQL_cloneConnection, 1},
+  {"RS_PostgreSQL_closeConnection", (DL_FUNC) &RS_PostgreSQL_closeConnection, 1},
+  {"RS_PostgreSQL_closeManager", (DL_FUNC) &RS_PostgreSQL_closeManager, 1},
+  {"RS_PostgreSQL_closeResultSet", (DL_FUNC) &RS_PostgreSQL_closeResultSet, 1},
+  {"RS_PostgreSQL_connectionInfo", (DL_FUNC) &RS_PostgreSQL_connectionInfo, 1},
+  {"RS_PostgreSQL_CopyIn", (DL_FUNC) &RS_PostgreSQL_CopyIn, 2},
+  {"RS_PostgreSQL_CopyInDataframe", (DL_FUNC) &RS_PostgreSQL_CopyInDataframe, 4},
+  {"RS_PostgreSQL_dbApply", (DL_FUNC) &RS_PostgreSQL_dbApply, 6},
+  {"RS_PostgreSQL_escape", (DL_FUNC) &RS_PostgreSQL_escape, 2},
+  {"RS_PostgreSQL_escape_bytea", (DL_FUNC) &RS_PostgreSQL_escape_bytea, 2},
+  {"RS_PostgreSQL_unescape_bytea", (DL_FUNC) &RS_PostgreSQL_unescape_bytea, 1},
+  {"RS_PostgreSQL_exec", (DL_FUNC) &RS_PostgreSQL_exec, 2},
+  {"RS_PostgreSQL_fetch", (DL_FUNC) &RS_PostgreSQL_fetch, 2},
+  {"RS_PostgreSQL_getException", (DL_FUNC) &RS_PostgreSQL_getException, 1},
+  {"RS_PostgreSQL_getResult", (DL_FUNC) &RS_PostgreSQL_getResult, 1},
+  {"RS_PostgreSQL_init", (DL_FUNC) &RS_PostgreSQL_init, 2},
+  {"RS_PostgreSQL_managerInfo", (DL_FUNC) &RS_PostgreSQL_managerInfo, 1},
+  {"RS_PostgreSQL_newConnection", (DL_FUNC) &RS_PostgreSQL_newConnection, 2},
+  {"RS_PostgreSQL_pqexec", (DL_FUNC) &RS_PostgreSQL_pqexec, 2},
+  {"RS_PostgreSQL_resultSetInfo", (DL_FUNC) &RS_PostgreSQL_resultSetInfo, 1},
+  {"RS_PostgreSQL_typeNames", (DL_FUNC) &RS_PostgreSQL_typeNames, 1},
+  {NULL, NULL, 0},
+};
+
+static const R_ExternalMethodDef extMethods[] = {
+  {"RS_PostgreSQL_pqexecparams", (DL_FUNC) &RS_PostgreSQL_pqexecparams, -1},
+  {NULL, NULL, 0},
+};
+
+
+void
+R_init_RPostgreSQL(DllInfo *info)
+{
+  R_registerRoutines(info, cMethods, callMethods, NULL, extMethods);
+  R_useDynamicSymbols(info, FALSE);
+  R_forceSymbols(info, TRUE);
+}
+
+
+
 Mgr_Handle *
 RS_PostgreSQL_init(s_object * config_params, s_object * reload)
 {
@@ -113,7 +196,6 @@ RS_PostgreSQL_closeManager(Mgr_Handle * mgrHandle)
 }
 
 
-
 /* open a connection with the same parameters used for in
  *  conHandle
  */
@@ -129,7 +211,7 @@ RS_PostgreSQL_cloneConnection(Con_Handle * conHandle)
     con = RS_DBI_getConnection(conHandle);
     conParams = con->conParams;
 
-    mgrHandle = RS_DBI_asMgrHandle(MGR_ID(conHandle));
+    PROTECT(mgrHandle = RS_DBI_asMgrHandle(MGR_ID(conHandle)));
 
 
     /* Connection parameters need to be put into a 8-element character
@@ -146,6 +228,7 @@ RS_PostgreSQL_cloneConnection(Con_Handle * conHandle)
     SET_CHR_EL(con_params, 6, C_S_CPY(conParams->options));
 
     MEM_UNPROTECT(1);
+    UNPROTECT(1);
 
     return RS_PostgreSQL_newConnection(mgrHandle, con_params);
 }
@@ -205,44 +288,22 @@ RS_PostgreSQL_newConnection(Mgr_Handle * mgrHandle, s_object * con_params)
     Con_Handle *conHandle;
     PGconn *my_connection;
 
-    char *user = NULL, *password = NULL, *host = NULL, *dbname = NULL, *port = NULL, *tty = NULL, *options = NULL;
+    const char *user = NULL, *password = NULL, *host = NULL, *dbname = NULL, *port = NULL, *tty = NULL, *options = NULL;
 
     if (!is_validHandle(mgrHandle, MGR_HANDLE_TYPE)) {
         RS_DBI_errorMessage("invalid PostgreSQLManager", RS_DBI_ERROR);
     }
 
-#define IS_EMPTY(s1)   !strcmp((s1), "")
-
-    if (!IS_EMPTY(CHR_EL(con_params, 0))) {
-        user = (char *) CHR_EL(con_params, 0);
-    }
-    if (!IS_EMPTY(CHR_EL(con_params, 1))) {
-        password = (char *) CHR_EL(con_params, 1);
-    }
-    if (!IS_EMPTY(CHR_EL(con_params, 2))) {
-        host = (char *) CHR_EL(con_params, 2);
-    }
-    if (!IS_EMPTY(CHR_EL(con_params, 3))) {
-        dbname = (char *) CHR_EL(con_params, 3);
-    }
-    if (!IS_EMPTY(CHR_EL(con_params, 4))) {
-        port = (char *) CHR_EL(con_params, 4);
-    }
-    if (!IS_EMPTY(CHR_EL(con_params, 5))) {
-        tty = (char *) CHR_EL(con_params, 5);
-    }
-    if (!IS_EMPTY(CHR_EL(con_params, 6))) {
-        options = (char *) CHR_EL(con_params, 6);
-    }
+    user = CHR_EL(con_params, 0);
+    password = CHR_EL(con_params, 1);
+    host = CHR_EL(con_params, 2);
+    dbname = CHR_EL(con_params, 3);
+    port = CHR_EL(con_params, 4);
+    tty = CHR_EL(con_params, 5);
+    options = CHR_EL(con_params, 6);
 
     my_connection = PQsetdbLogin(host, port, options, tty, dbname, user, password);
 
-    if (PQstatus(my_connection) != CONNECTION_OK) {
-        char buf[1000];
-	sprintf(buf, "could not connect %s@%s on dbname \"%s\"\n", PQuser(my_connection), host?host:"local", PQdb(my_connection));
-        RS_DBI_errorMessage(buf, RS_DBI_ERROR);
-    }
-
     conParams = RS_postgresql_allocConParams();
 
     /* save actual connection parameters */
@@ -261,16 +322,33 @@ RS_PostgreSQL_newConnection(Mgr_Handle * mgrHandle, s_object * con_params)
     conParams->tty = RS_DBI_copyString(PQtty(my_connection));
     conParams->options = RS_DBI_copyString(PQoptions(my_connection));
 
+    if (PQstatus(my_connection) != CONNECTION_OK) {
+        char buf[1024];
+	snprintf(buf, 1023, 
+           "could not connect %s@%s:%s on dbname \"%s\": %s", 
+           conParams->user, conParams->host, conParams->port,
+           conParams->dbname, PQerrorMessage(my_connection));
+        buf[1023]= '\0';
+        PQfinish(my_connection);
+        my_connection = NULL;
+        RS_PostgreSQL_freeConParams(conParams); /*free BEFORE emitting err message that do not come back */
+        RS_DBI_errorMessage(buf, RS_DBI_ERROR);
+        return R_NilValue; /* don't reach here as it goes back to R proc */
+    }
+
     PROTECT(conHandle = RS_DBI_allocConnection(mgrHandle, (Sint) 1)); /* The second argument (1) specifies the number of result sets allocated */
     con = RS_DBI_getConnection(conHandle);
-    if (!con) {
+    if (my_connection && !con) {
         PQfinish(my_connection);
+        my_connection = NULL;
         RS_PostgreSQL_freeConParams(conParams);
         conParams = (RS_PostgreSQL_conParams *) NULL;
         RS_DBI_errorMessage("could not alloc space for connection object", RS_DBI_ERROR);
     }
-    con->drvConnection = (void *) my_connection;
-    con->conParams = (void *) conParams;
+    if(con){
+        con->drvConnection = (void *) my_connection;
+        con->conParams = (void *) conParams;
+    }
     UNPROTECT(1);
     return conHandle;
 }
@@ -357,10 +435,9 @@ RS_PostgreSQL_exec(Con_Handle * conHandle, s_object * statement)
         omsg = PQerrorMessage(my_connection);
         len = strlen(omsg);
         free(dyn_statement);
-        errMsg = malloc(len + 80); /* 80 should be larger than the length of "could not ..."*/
+        errMsg = R_alloc(len + 80, 1); /* 80 should be larger than the length of "could not ..."*/
         snprintf(errMsg, len + 80,  "could not run statement: %s", omsg);
         RS_DBI_errorMessage(errMsg, RS_DBI_ERROR);
-        free(errMsg);
     }
 
 
@@ -376,23 +453,18 @@ RS_PostgreSQL_exec(Con_Handle * conHandle, s_object * statement)
     /* char *PQresultErrorMessage(const PGresult *res); */
 
     if (strcmp(PQresultErrorMessage(my_result), "") != 0) {
-
-        free(dyn_statement);
         char *errResultMsg;
         const char *omsg;
         size_t len;
         omsg = PQerrorMessage(my_connection);
         len = strlen(omsg);
-        errResultMsg = malloc(len + 80); /* 80 should be larger than the length of "could not ..."*/
+        errResultMsg = R_alloc(len + 80, 1); /* 80 should be larger than the length of "could not ..."*/
         snprintf(errResultMsg, len + 80, "could not Retrieve the result : %s", omsg);
-        RS_DBI_errorMessage(errResultMsg, RS_DBI_ERROR);
-        free(errResultMsg);
-
         /*  Frees the storage associated with a PGresult.
          *  void PQclear(PGresult *res);   */
-
         PQclear(my_result);
-
+        free(dyn_statement);
+        RS_DBI_errorMessage(errResultMsg, RS_DBI_ERROR);
     }
 
     /* we now create the wrapper and copy values */
@@ -443,7 +515,7 @@ RS_PostgreSQL_createDataMappings(Res_Handle * rsHandle)
     con = RS_DBI_getConnection(rsHandle);
     num_fields = PQnfields(my_result);
 
-    flds = RS_DBI_allocFields(num_fields);
+    flds = RS_DBI_allocFields(num_fields); /* this returns malloced data (not from R) */
 
     char buff[1000];            /* Buffer to hold the sql query to check whether the given column is nullable */
     PGconn *conn;
@@ -587,18 +659,15 @@ RS_PostgreSQL_fetch(s_object * rsHandle, s_object * max_rec)
     RS_DBI_fields *flds;
     PGresult *my_result;
     s_object *output, *s_tmp;
-#ifndef USING_R
-    s_object *raw_obj, *raw_container;
-#endif
-    /* unsigned long  *lens;  */
     int i, j, null_item, expand;
     Sint completed;
-    /* Sint * fld_nullOk set but not used */
     Stype *fld_Sclass;
     Sint num_rec;
     int num_fields;
     int num_rows;               /*num_rows added to count number of rows */
     int k;                      /* This takes care of pointer to the required row */
+    /* Encoding */
+    cetype_t r_encoding = CE_NATIVE;
 
     result = RS_DBI_getResultSet(rsHandle);
     flds = result->fields;
@@ -620,8 +689,6 @@ RS_PostgreSQL_fetch(s_object * rsHandle, s_object * max_rec)
     MEM_PROTECT(output = NEW_LIST((Sint) num_fields));
     RS_DBI_allocOutput(output, flds, num_rec, 0);
     fld_Sclass = flds->Sclass;
-/*    fld_nullOk = flds->nullOk; set but not used */
-
     /* actual fetching.... */
     my_result = (PGresult *) result->drvResultSet;
 
@@ -629,6 +696,22 @@ RS_PostgreSQL_fetch(s_object * rsHandle, s_object * max_rec)
 
     k = result->rowCount;       /* ADDED */
 
+
+    if(k < num_rows){
+      const char * pq_enc_string;
+      int pq_encoding = 0;
+      RS_DBI_connection *con;
+      con = RS_DBI_getConnection(rsHandle);
+      pq_encoding = PQclientEncoding((PGconn *) con->drvConnection);
+      pq_enc_string = pg_encoding_to_char(pq_encoding);
+      if(strcmp(pq_enc_string, "UTF8") == 0){
+        r_encoding = CE_UTF8;
+      }
+      if(strcmp(pq_enc_string, "LATIN1") == 0){
+        r_encoding = CE_LATIN1;
+      }
+    }
+
     completed = (Sint) 0;
     for (i = 0;; i++, k++) {
         if (k >= num_rows) {
@@ -700,14 +783,10 @@ RS_PostgreSQL_fetch(s_object * rsHandle, s_object * max_rec)
 
             case CHARACTER_TYPE:
                 if (null_item) {
-#ifdef USING_R
                     SET_LST_CHR_EL(output, j, i, NA_STRING);
-#else
-                    NA_CHR_SET(LST_CHR_EL(output, j, i));
-#endif
                 }
                 else {
-                    SET_LST_CHR_EL(output, j, i, C_S_CPY(PQgetvalue(my_result, k, j)));
+                    SET_LST_CHR_EL(output, j, i, mkCharCE(PQgetvalue(my_result, k, j), r_encoding));
                 }
                 break;
 
@@ -720,31 +799,9 @@ RS_PostgreSQL_fetch(s_object * rsHandle, s_object * max_rec)
                 }
                 break;
 
-#ifndef USING_R
-            case SINGLE_TYPE:
-                if (null_item) {
-                    NA_SET(&(LST_FLT_EL(output, j, i)), SINGLE_TYPE);
-                }
-                else {
-                    LST_FLT_EL(output, j, i) = (float) atof(PQgetvalue(my_result, k, j));
-                }
-                break;
-
-            case RAW_TYPE:     /* these are blob's */
-                raw_obj = NEW_RAW((Sint) PQgetlength(my_result, k, j));
-                memcpy(RAW_DATA(raw_obj), PQgetvalue(my_result, k, j), PQgetlength(my_result, k, j));
-                raw_container = LST_EL(output, j);
-                SET_ELEMENT(raw_container, (Sint) i, raw_obj);
-                SET_ELEMENT(output, (Sint) j, raw_container);
-                break;
-#endif
             default:
                 if (null_item) {
-#ifdef USING_R
                     SET_LST_CHR_EL(output, j, i, NA_STRING);
-#else
-                    NA_CHR_SET(LST_CHR_EL(output, j, i));
-#endif
                 }
                 else {
                     char warn[64];
@@ -899,54 +956,45 @@ RS_PostgreSQL_connectionInfo(Con_Handle * conHandle)
     RS_PostgreSQL_conParams *conParams;
     RS_DBI_connection *con;
     s_object *output;
-    Sint i, n = 7 /*8 */ , *res, nres;
-    char *conDesc[] = { "host", "user", "dbname",
+    Sint i, n = 8, *res, nres;
+    int sv, major, minor_revision, minor, revision_num;
+    char *conDesc[] = { "host", "port", "user", "dbname",
         "serverVersion", "protocolVersion",
         "backendPId", "rsId"
     };
     Stype conType[] = { CHARACTER_TYPE, CHARACTER_TYPE, CHARACTER_TYPE,
-        /* CHARACTER_TYPE, */ CHARACTER_TYPE, INTEGER_TYPE,
+        CHARACTER_TYPE, CHARACTER_TYPE, INTEGER_TYPE,
         INTEGER_TYPE, INTEGER_TYPE
     };
-    Sint conLen[] = { 1, 1, 1 /*, 1 */ , 1, 1, 1, 1 };
+    Sint conLen[] = { 1, 1, 1, 1, 1, 1, 1, -1 };
 
     con = RS_DBI_getConnection(conHandle);
-    conLen[6 /*7 */ ] = con->num_res;   /* num of open resultSets */
-    my_con = (PGconn *) con->drvConnection;
+    conLen[7] = con->num_res;   
     PROTECT(output = RS_DBI_createNamedList(conDesc, conType, conLen, n));
     conParams = (RS_PostgreSQL_conParams *) con->conParams;
 
     SET_LST_CHR_EL(output, 0, 0, C_S_CPY(conParams->host));
-    SET_LST_CHR_EL(output, 1, 0, C_S_CPY(conParams->user));
-    SET_LST_CHR_EL(output, 2, 0, C_S_CPY(conParams->dbname));
+    SET_LST_CHR_EL(output, 1, 0, C_S_CPY(conParams->port));
+    SET_LST_CHR_EL(output, 2, 0, C_S_CPY(conParams->user));
+    SET_LST_CHR_EL(output, 3, 0, C_S_CPY(conParams->dbname));
 
-    /* PQserverVersion: Returns an integer representing the backend version.
-     * Syntax: int PQserverVersion(const PGconn *conn);
-     */
-    /*Long int is taken because in some Operating sys int is 2 bytes which will not be enough for     * server version number
-     */
-    long int sv = PQserverVersion(my_con);
+    my_con = (PGconn *) con->drvConnection;
+    sv = PQserverVersion(my_con);
 
-    int major = (int) sv / 10000;
-    int minor = (int) (sv % 10000) / 100;
-    int revision_num = (int) (sv % 10000) % 100;
+    major = sv / 10000;
+    minor_revision = sv % 10000;
+    minor = minor_revision / 100;
+    revision_num = minor_revision % 100;
 
-    char buf1[50];
+    {
+        char buf1[64];
+        snprintf(buf1, 63, "%d.%d.%d", major, minor, revision_num);
+        buf1[63]='\0';
+        SET_LST_CHR_EL(output, 4, 0, C_S_CPY(buf1));
+    }
 
-    sprintf(buf1, "%d.%d.%d", major, minor, revision_num);
-
-    SET_LST_CHR_EL(output, 3, 0, C_S_CPY(buf1));
-
-    /* PQprotocolVersion: Interrogates the frontend/backend protocol being used.
-     * int PQprotocolVersion(const PGconn *conn);
-     */
-    LST_INT_EL(output, 4 /*5 */ , 0) = (Sint) PQprotocolVersion(my_con);
-
-    /* PQbackendPID: Returns the process ID (PID) of the backend server process handling     * this connection.
-     * Syntax: int PQbackendPID(const PGconn *conn);
-     */
-
-    LST_INT_EL(output, 5 /*6 */ , 0) = (Sint) PQbackendPID(my_con);
+    LST_INT_EL(output, 5, 0) = (Sint) PQprotocolVersion(my_con);
+    LST_INT_EL(output, 6, 0) = (Sint) PQbackendPID(my_con);
 
     res = (Sint *) S_alloc((long) con->length, (int) sizeof(Sint));
     nres = RS_DBI_listEntries(con->resultSetIds, con->length, res);
@@ -955,7 +1003,7 @@ RS_PostgreSQL_connectionInfo(Con_Handle * conHandle)
     }
 
     for (i = 0; i < con->num_res; i++) {
-        LST_INT_EL(output, 6, i) = (Sint) res[i];
+        LST_INT_EL(output, 7, i) = (Sint) res[i];
     }
     UNPROTECT(1);
     return output;
